@@ -83,7 +83,7 @@ class Model:
         if len(ic.shape) == 1:
             self.dimension = 0
             self.N = 0
-        if phys_domain is None and self.dimension is not 0:
+        if phys_domain is None and self.dimension != 0:
             self.domain = [(0, self.N) for i in range(self.dimension)]
         else:
             # assert self.dimension is not 0, "Cannot specify physical domain for ODE system."
@@ -106,18 +106,22 @@ class Model:
         self.sigma = None
         self.lib_strings = None
 
-    def init_library(self, ode_order, pde_order, indices=None, custom_terms=np.array([]), kind="split"):
+        self.sol_dot = None
+        self.lot = None
+
+    def init_library(self, ode_order=2, pde_order=2, indices=None, custom_terms=np.array([]), kind="split"):
         """
         Method that initializes all library strings up to specified order. These string will be pre-compiled and later
         evaluated in the right hand side method. They include products of all provided variabled up to order ode_order
         and their spatial derivatives up to order pde_order. Indices is used to filter only selected library terms for
         spatial differentiation.
+
         Parameters
         ----------
         ode_order : int
-            Order up to which products of variables (including with themselves) are included in the library
+            (Optional, default=2) Order up to which products of variables (including with themselves) are included in the library
         pde_order : int
-            Order up to which spatial derivatives are to be computed
+            (Optional, default=2) Order up to which spatial derivatives are to be computed
         indices : np.array
             (optional, default=None) Array of indices of library elements to be included for spatial derivation (as a
             subset of terms created undeer ode_order)
@@ -131,9 +135,12 @@ class Model:
         -----
         Library creation should focus on the array-valued attribute "lib_string" and later evaluation should use that
         array for easier manipulation of library terms. Then one wouldn't need all this bogus differentiation between
-        ode, pde or custom strings.
+        ode, pde or custom strings. Also using strings may generally be prohibitive.
 
         """
+        if self.dimension == 0:
+            pde_order = 0
+
         self.ode_order = ode_order  # Maximum number of factors in products of permutations
         self.pde_order = pde_order  # Maximum order of spatial derivative
         self.ode_count = self._give_count(self.ode_order)
@@ -168,7 +175,7 @@ class Model:
                             if comb[0] != comb[1]:
                                 x_string += f"self.diff.dd_dxdy({ode_strings[index]},[{comb[1]},{comb[0]}]), "
                                 self.pde_count += 1
-            x_string = x_string[:-2] + (pde_order == 0)*"([" + "])"
+            x_string = x_string[:-2] + (pde_order == 0) * "([" + "])"
             return x_string
 
         def make_coupled_pde_terms():
@@ -334,9 +341,6 @@ class Model:
         time : numpy.array
             (optional, default=None) Time stepping data corresponding to solution, if None self.time is used.
 
-        Yields
-        -------
-        TimeSimulator object in Model attributes
         """
         if sig is None:
             sigma = self.sigma
@@ -354,6 +358,7 @@ class Model:
         domain = self.domain
         boundary_condition = self.boundary_condition
         differentiation_method = self.differentiation_method
+
         if self.dimension == 1:
             n_shape = (self.N,)
         elif self.dimension == 2:
@@ -367,14 +372,14 @@ class Model:
                 self.library_over_time = None
                 self.sol_dot = None
 
-            def rhs(self, t, u):
-                ode_lib = eval(ode_string)
-                pde_lib = eval(pde_string)
-                custom_lib = eval(custom_string)
-                vec = np.concatenate((ode_lib, pde_lib, custom_lib))
-                u_next = [np.dot(s, [item.flatten() for item in vec]).reshape(n_shape) for s in self.sigma]
-
-                return np.array(u_next)
+            # def rhs(self, t, u):
+            #     ode_lib = eval(ode_string)
+            #     pde_lib = eval(pde_string)
+            #     custom_lib = eval(custom_string)
+            #     vec = np.concatenate((ode_lib, pde_lib, custom_lib))
+            #     u_next = [np.dot(s, [item.flatten() for item in vec]).reshape(n_shape) for s in self.sigma]
+            #
+            #     return np.array(u_next)
 
             def create_library_time_series(self, ts=None):
                 if ts is None:
@@ -400,7 +405,30 @@ class Model:
                 self.library_over_time = np.array(library_over_time)
                 return self.library_over_time
 
-        self.simulator = TimeSimulator()
+        def rhs_pde(obj, t, u):
+            ode_lib = eval(ode_string)
+            pde_lib = eval(pde_string)
+            custom_lib = eval(custom_string)
+            vec = np.concatenate((ode_lib, pde_lib, custom_lib))
+            u_next = [np.dot(s, [item.flatten() for item in vec]).reshape(n_shape) for s in self.sigma]
+            return np.array(u_next)
+
+        def rhs_ode(obj, t, u):
+            ode_lib = eval(ode_string)
+            pde_lib = eval(pde_string)
+            custom_lib = eval(custom_string)
+            vec = np.concatenate((ode_lib, pde_lib, custom_lib))
+            u_next = [np.dot(s, [item for item in vec]) for s in self.sigma]
+            return np.array(u_next)
+
+        # This is actually batshit crazy
+        sim = TimeSimulator()
+        if self.dimension == 0:
+            sim.rhs = rhs_ode.__get__(sim)
+        else:
+            sim.rhs = rhs_pde.__get__(sim)
+
+        self.simulator = sim
 
         # Usually doesn't do anything
         self.simulator.sol = sol
@@ -426,9 +454,13 @@ class Model:
                 sigma_loc = self.sigma
             headers = ["Index", "Term"]
             for i in range(sigma_loc.shape[0]):
-                headers.append(f"Value {i}")
+                headers.append(f"Value {i}")  # could do this with list comprehension
 
-            print(tabulate(zip(range(len(self.lib_strings)), self.lib_strings, *sigma_loc), headers=headers))
+            strings = [item[10:] if item[:10] == "self.diff." else item for item in self.lib_strings]
+            if strings[0] == "np.ones(u[0].shape)":
+                strings[0] = "1.0"
+
+            print(tabulate(zip(range(len(self.lib_strings)), strings, *sigma_loc), headers=headers))
 
     def optimize_sigma(self, lamb=0.1, thres=0.01, error="BIC", **kwargs):
         """
@@ -444,19 +476,54 @@ class Model:
             (Optional, default=0.01) Threshold below which values are cut in sequence of regression
 
         error : string
-            (Optional, default="BIC") Keyword argument to pass to optimization routine. Can either use BIC with
-            integrated error or non-integrated error otherwise.
+            (Optional, default="BIC") Keyword argument to pass to optimization routine. Choices are "BIC" for
+            bayesian information criterion using integration, "SINDy" for the sum of squared differences in derivative
+            and library reconstruction and "integrate" for sum of squared differences in original time series and
+            integrated library reconstruction. (Both "SINDy" and "integrate" additionally punish non-sparsity,
+            adjustable with the additional keyword argument `l0`.)
 
         """
         print("Generating test functions (this may take some time)...")
-        sol_dot = self.simulator.get_sol_dot()
-        lot = self.simulator.create_library_time_series()
+        if self.sol_dot is None:
+            self.sol_dot = self.simulator.get_sol_dot()
+        if self.lot is None:
+            self.lot = self.simulator.create_library_time_series()
         # sol_dot_flat = sol_dot.reshape((sol_dot.shape[0], -1))
         # lot_flat = lot.reshape((lot.shape[0], lot.shape[1], -1))
 
-        sigma_opt, error = train_ridge_wip(lot, self.sol, sol_dot, self.time, lamb, thres, error=error, **kwargs)
+        sigma_opt, error = train_ridge_wip(self.lot, self.sol, self.sol_dot, self.time, lamb, thres, error=error,
+                                           **kwargs)
 
         self.sigma = sigma_opt
         print("New Sigma set to: \n")
 
         self.print_library()
+
+    def least_square_sigma(self, lamb=None):
+        """
+        Method that performs a least-square fit using numpy to fit the library to the derivative of the time series,
+        without any regard for sparsity or further optimization (mainly for debugging purposes).
+
+        Parameters
+        ----------
+        lamb : float or None
+            (Optional, default=None) If provided adds the ridge-regression term for promoting sparsity, then `lamb` is
+            the sparsity knob otherwise perform ordinary least-square fit
+
+        """
+        if self.sol_dot is None:
+            self.sol_dot = self.simulator.get_sol_dot()
+        if self.lot is None:
+            self.lot = self.simulator.create_library_time_series()
+
+        sol_dot_flat = self.sol_dot.reshape((self.sol_dot.shape[0], -1))
+        lot_flat = self.lot.reshape((self.lot.shape[0], self.lot.shape[1], -1))
+        if lamb is None:
+            sigma = [np.linalg.lstsq(lot_flat[i].T, sol_dot_flat[i], rcond=None)[0] for i in range(self.n_variables)]
+        else:
+            sigma = [np.linalg.lstsq(lot_flat[i].dot(lot_flat[i].T) +
+                                     lamb * np.eye(self.lot.shape[1]), lot_flat[i].dot(sol_dot_flat[i]), rcond=None)[0]
+                     for i in range(self.n_variables)]
+        self.sigma = np.array(sigma)
+
+        self.print_library()
